# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/TimeToLiveApi.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.ai/h2o/featurestore/api/v1/TimeToLiveApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"L\n\x06Online\"B\n\x12TimeToLiveInterval\x12\x08\n\x04\x44\x41YS\x10\x00\x12\n\n\x06MONTHS\x10\x01\x12\t\n\x05HOURS\x10\x02\x12\x0b\n\x07MINUTES\x10\x03\"K\n\x07Offline\"@\n\x12TimeToLiveInterval\x12\t\n\x05YEARS\x10\x00\x12\x08\n\x04\x44\x41YS\x10\x01\x12\n\n\x06MONTHS\x10\x02\x12\t\n\x05HOURS\x10\x03\"\xdf\x01\n\nTimeToLive\x12\x13\n\x0bttl_offline\x18\x01 \x01(\r\x12T\n\x14ttl_offline_interval\x18\x02 \x01(\x0e\x32\x36.ai.h2o.featurestore.api.v1.Offline.TimeToLiveInterval\x12\x12\n\nttl_online\x18\x03 \x01(\r\x12R\n\x13ttl_online_interval\x18\x04 \x01(\x0e\x32\x35.ai.h2o.featurestore.api.v1.Online.TimeToLiveIntervalB\x14\x42\x12TimeToLiveApiProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ai.h2o.featurestore.api.v1.TimeToLiveApi_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'B\022TimeToLiveApiProto'
  _globals['_ONLINE']._serialized_start=78
  _globals['_ONLINE']._serialized_end=154
  _globals['_ONLINE_TIMETOLIVEINTERVAL']._serialized_start=88
  _globals['_ONLINE_TIMETOLIVEINTERVAL']._serialized_end=154
  _globals['_OFFLINE']._serialized_start=156
  _globals['_OFFLINE']._serialized_end=231
  _globals['_OFFLINE_TIMETOLIVEINTERVAL']._serialized_start=167
  _globals['_OFFLINE_TIMETOLIVEINTERVAL']._serialized_end=231
  _globals['_TIMETOLIVE']._serialized_start=234
  _globals['_TIMETOLIVE']._serialized_end=457
# @@protoc_insertion_point(module_scope)



import time

import grpc
import pkg_resources
import requests
from google.protobuf.empty_pb2 import Empty
from grpc import ChannelConnectivity
from requests.packages.urllib3.exceptions import InsecureRequestWarning

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1 import CoreService_pb2_grpc
from featurestore.core import interactive_console
from featurestore.core.acl import AccessControlList
from featurestore.core.auth import AuthWrapper
from featurestore.core.client_config import ClientConfig
from featurestore.core.collections.classifiers import Classifiers
from featurestore.core.collections.jobs import Jobs
from featurestore.core.collections.projects import Projects
from featurestore.core.collections.transformation_functions import TransformationFunctions
from featurestore.core.config import ConfigUtils
from featurestore.core.credentials import CredentialsHelper
from featurestore.core.data_source_wrappers import SparkDataFrame, get_raw_data_location
from featurestore.core.entities.extract_schema_job import ExtractSchemaJob
from featurestore.core.interceptors import AuthClientInterceptor, ExponentialBackoff, RetryOnRpcErrorClientInterceptor
from featurestore.core.schema import Schema


class Client:
    def __init__(
        self,
        url: str,
        secure: bool = False,
        root_certificates: str = None,
        config: ClientConfig = ClientConfig(),
    ):
        self._client_config = config
        options = [
            ("grpc.primary_user_agent", "feature-store-py-cli/SUBST_FS_VERSION"),
        ]
        if secure:
            default_root_certs = pkg_resources.resource_filename("featurestore", "_credentials/roots.pem")
            credentials = self._get_channel_credentials(root_certificates if root_certificates else default_root_certs)
            channel = grpc.secure_channel(url, credentials, options)
        else:
            channel = grpc.insecure_channel(url, options)

        interceptors = [
            RetryOnRpcErrorClientInterceptor(
                max_attempts=5,
                sleeping_policy=ExponentialBackoff(init_backoff_ms=1000, max_backoff_ms=30000, multiplier=4),
                status_for_retry=(
                    grpc.StatusCode.UNAVAILABLE,
                    grpc.StatusCode.DEADLINE_EXCEEDED,
                ),
            ),
            AuthClientInterceptor(self),
        ]
        self._channel = grpc.intercept_channel(channel, *interceptors)
        self._connection_state = None
        self._config = ConfigUtils.collect_properties()

        def on_connectivity_change(value):
            self._connection_state = value
            return

        self._channel.subscribe(on_connectivity_change, try_to_connect=True)
        if config.wait_for_backend:
            while self._connection_state != ChannelConnectivity.READY:
                print("Connecting to the server {} ...".format(url))
                time.sleep(2)
        else:
            print("Connected to server {}".format(url))

        self._stub = CoreService_pb2_grpc.CoreServiceStub(self._channel)
        self.auth = AuthWrapper(self._stub)
        self.projects = Projects(self._stub)
        self.jobs = Jobs(self._stub)
        self.classifiers = Classifiers(self._stub)
        self.transformation_functions = TransformationFunctions(self._stub)
        self.acl = AccessControlList(self._stub)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._channel.close()
        return False

    def close(self):
        self._channel.close()

    def get_version(self) -> str:
        request = Empty()
        response = self._stub.GetVersion(request)
        return response.version

    def extract_schema_from_source_async(self, raw_data_location, credentials=None) -> ExtractSchemaJob:
        request = pb.StartExtractSchemaJobRequest()
        if isinstance(raw_data_location, SparkDataFrame):
            raw_data_location._write_to_cache(self._stub)
            data_source = raw_data_location._get_cache_location()
        else:
            data_source = get_raw_data_location(raw_data_location)
        request.raw_data.CopyFrom(data_source)
        CredentialsHelper.set_credentials(request, data_source, credentials)
        job_id = self._stub.StartExtractSchemaJob(request)
        return ExtractSchemaJob(self._stub, job_id)

    @interactive_console.record_stats
    def extract_schema_from_source(self, raw_data_location, credentials=None) -> Schema:
        job = self.extract_schema_from_source_async(raw_data_location, credentials)
        return job.wait_for_result()

    @interactive_console.record_stats
    def extract_derived_schema(self, feature_sets, transformation) -> Schema:
        job = self.extract_derived_schema_async(feature_sets, transformation)
        return job.wait_for_result()

    def extract_derived_schema_async(self, feature_sets, transformation) -> Schema:
        transformation._initialize(self._stub)
        request = pb.StartExtractSchemaJobRequest(
            derived_from=pb.DerivedInformation(
                feature_set_ids=[pb.VersionedId(id=f.id, major_version=f.major_version) for f in feature_sets],
                transformation=transformation._to_proto(),
            )
        )

        job_id = self._stub.StartExtractSchemaJob(request)
        return ExtractSchemaJob(self._stub, job_id)

    def _has_online_retrieve_permissions(self, project_name, feature_set_name):
        request = pb.HasPermissionToRetrieveRequest()
        request.project_name = project_name
        request.feature_set_name = feature_set_name
        response = self._stub.HasPermissionToRetrieve(request)
        return response.has_retrieve_permission

    def online_retrieve(
        self,
        project_name,
        feature_set_name,
        primary_key,
        secondary_key=None,
        verify=False,
    ):
        # this method will be eventually removed, H2O's online is using method onlineRetrieve on feature set entity
        if not self._has_online_retrieve_permissions(project_name, feature_set_name):
            raise Exception("User does not have enough permissions to execute online retrieve call")
        resp = self._stub.GetOnlineRetrieveMeta(Empty())
        if not resp.is_enabled:
            raise Exception("Online feature store integration is not enabled on the backend")
        url = (
            resp.base_online_url
            + requests.utils.quote(project_name)
            + "/"
            + requests.utils.quote(feature_set_name)
            + "/"
            + requests.utils.quote(primary_key)
        )
        if secondary_key:
            params = dict(secondaryKey=requests.utils.quote(secondary_key))
        else:
            params = dict(secondaryKey="")
        session = requests.Session()
        session.auth = (resp.username, resp.password)
        session.headers.update({"user-id": self.auth.get_active_user().id})
        if not verify:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        response = session.get(url, params=params, verify=verify).json()
        return response

    def show_progress(self, interactive):
        ConfigUtils.set_property(self._config, ConfigUtils.INTERACTIVE_LOGGING, str(interactive))

    @staticmethod
    def _get_channel_credentials(cert_location: str) -> grpc.ChannelCredentials:
        if cert_location is not None:
            with open(cert_location, "rb") as cert_file:
                return grpc.ssl_channel_credentials(cert_file.read())
        return grpc.ssl_channel_credentials(None)



import json

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..utils import Utils


class ScheduledTask:
    def __init__(self, stub, scheduled_task):
        st = pb.ScheduledTask()
        st.CopyFrom(scheduled_task)
        self._scheduled_task = st
        self._stub = stub

    @property
    def id(self):
        return self._scheduled_task.id

    @property
    def name(self):
        return self._scheduled_task.name

    @property
    def description(self):
        return self._scheduled_task.description

    @description.setter
    def description(self, value: str):
        update_request = pb.ScheduledTaskUpdateRequest(
            scheduled_task_id=self._scheduled_task.id,
            description=value,
            updated_fields=["description"],
        )
        self._stub.UpdateScheduledTask(update_request)
        self._refresh()

    @property
    def schedule(self):
        return self._scheduled_task.schedule

    @schedule.setter
    def schedule(self, value: str):
        update_request = pb.ScheduledTaskUpdateRequest(
            scheduled_task_id=self._scheduled_task.id,
            schedule=value,
            updated_fields=["schedule"],
        )
        self._stub.UpdateScheduledTask(update_request)
        self._refresh()

    def delete(self):
        request = pb.ScheduledTaskId(scheduled_task_id=self._scheduled_task.id)
        self._stub.DeleteScheduledTask(request)

    def _refresh(self):
        request = pb.ScheduledTaskId(scheduled_task_id=self._scheduled_task.id)
        response = self._stub.GetScheduledTask(request)
        self._scheduled_task = response.task

    def __repr__(self):
        last_exec = (
            ""
            if self._scheduled_task.last_exec.seconds == 0
            else Utils.convert_timestamp_to_str_with_zone(self._scheduled_task.last_exec)
        )
        last_update_time = (
            ""
            if self._scheduled_task.last_update_date_time.seconds == 0
            else Utils.convert_timestamp_to_str_with_zone(self._scheduled_task.last_update_date_time)
        )
        task_type = pb.TaskType.DESCRIPTOR.values_by_number.get(self._scheduled_task.task_type).name
        task = {
            "id": self._scheduled_task.id,
            "name": self._scheduled_task.name,
            "description": self._scheduled_task.description,
            "feature_set_id": self._scheduled_task.feature_set_id,
            "project_id": self._scheduled_task.project_id,
            "source": json.loads(Utils.pretty_print_proto(self._scheduled_task.source)),
            "schedule": self._scheduled_task.schedule,
            "next_run_time": Utils.convert_timestamp_to_str_with_zone(self._scheduled_task.next_run_time),
            "last_exec": last_exec,
            "owner": json.loads(Utils.pretty_print_proto(self._scheduled_task.owner)),
            "created_date_time": Utils.convert_timestamp_to_str_with_zone(self._scheduled_task.created_date_time),
            "last_update_date_time": last_update_time,
            "task_type": task_type,
            "feature_set_version": self._scheduled_task.feature_set_version,
        }
        return json.dumps(task, indent=2)

    def __str__(self):
        return (
            f"Name                  : {self.name} \n"
            f"Description           : {self.description} \n"
            f"Feature set version   : {self._scheduled_task.feature_set_version} \n"
            f"Owner                   \n{self._custom_user_string()}"
            f"Source                  \n{self._custom_source_string()}"
            f"Schedule              : {self.schedule} \n"
            f"Last execution        : {self._scheduled_task.last_exec} \n"
            f"Next run time         : {self._scheduled_task.next_run_time}"
        )

    def _custom_user_string(self):
        return (
            f"          Name        : {self._scheduled_task.owner.name} \n"
            f"          Email       : {self._scheduled_task.owner.email} \n"
        )

    def _custom_source_string(self):
        source_dict = Utils.proto_to_dict(self._scheduled_task.source)
        source = list(source_dict.keys())[0]
        tmp_str = ""
        for key, value in source_dict.get(source).items():
            tmp_str = tmp_str + Utils.output_indent_spacing(f"{key}: {value} \n", "           ")

        return (
            f"{Utils.output_indent_spacing(source, '       ')}: [ \n"
            f"{tmp_str}"
            f"{Utils.output_indent_spacing(']', '       ')} \n"
        )



import time

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..access_type import AccessType
from ..collections.feature_sets import FeatureSets
from ..collections.feature_views import FeatureViews
from ..utils import Utils
from .user import User


class Project:
    def __init__(self, stub, project):
        self._project = project
        self._stub = stub
        self.feature_sets = FeatureSets(stub, project)
        self.feature_views = FeatureViews(stub, project)

    @property
    def id(self):
        return self._project.id

    @property
    def name(self):
        return self._project.name

    @property
    def description(self):
        return self._project.description

    @description.setter
    def description(self, value):
        update_request = pb.UpdateProjectRequest(
            project_id=self._project.id, description=value, fields_to_update=[pb.PROJECT_DESCRIPTION]
        )
        self._project = self._stub.UpdateProject(update_request).updated_project

    @property
    def secret(self):
        return self._project.secret

    @secret.setter
    def secret(self, value):
        update_request = pb.UpdateProjectRequest(
            project_id=self._project.id, secret=value, fields_to_update=[pb.PROJECT_SECRET]
        )
        self._project = self._stub.UpdateProject(update_request).updated_project

    @property
    def locked(self):
        return self._project.locked

    @locked.setter
    def locked(self, value):
        update_request = pb.UpdateProjectRequest(
            project_id=self._project.id, locked=value, fields_to_update=[pb.PROJECT_LOCKED]
        )
        self._project = self._stub.UpdateProject(update_request).updated_project

    @property
    def owner(self):
        Utils.warn_deprecated('Owner field will be removed in 0.14.0. Please use "owners" instead.')
        return User(self._project.owner)

    @owner.setter
    def owner(self, email):
        Utils.warn_deprecated('Owner field will be removed in 0.14.0. Please use "owners" instead.')
        request = pb.GetUserByMailRequest()
        request.email = email
        response = self._stub.GetUserByMail(request)
        user = response.user
        update_request = pb.ProjectUserFieldUpdateRequest()
        update_request.new_value.CopyFrom(user)
        update_request.project_id = self._project.id
        self._stub.UpdateProjectOwner(update_request)
        self._refresh()

    @property
    def author(self):
        return User(self._project.author)

    @property
    def custom_data(self):
        return self._project.custom_data

    @custom_data.setter
    def custom_data(self, value):
        update_request = pb.UpdateProjectRequest(
            project_id=self._project.id, custom_data=value, fields_to_update=[pb.PROJECT_CUSTOM_DATA]
        )
        self._project = self._stub.UpdateProject(update_request).updated_project

    def delete(self, wait_for_completion=False):
        request = pb.DeleteProjectRequest()
        request.project.CopyFrom(self._project)
        self._stub.DeleteProject(request)
        exists_request = pb.ProjectExistsRequest()
        exists_request.project_id = self._project.id
        if wait_for_completion:
            while self._stub.ProjectExists(exists_request).exists:
                time.sleep(1)
                print("Waiting for project '{}' deletion".format(self._project.name))

    def add_owners(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.Owner)

    def add_editors(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.Editor)

    def add_consumers(self, user_emails):
        return self._add_permissions(user_emails, pb.PermissionType.Consumer)

    def add_sensitive_consumers(self, user_emails):
        """Add additional sensitive consumer/consumers to project.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            Project: An existing project with the latest information.

        Typical example:
            project.add_sensitive_consumers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#project-permission-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.SensitiveConsumer)

    def add_viewers(self, user_emails):
        """Add additional viewer/viewers to project.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            Project: An existing project with the latest information.

        Typical example:
            project.add_viewers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#project-permission-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.Viewer)

    def remove_owners(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.Owner)

    def remove_editors(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.Editor)

    def remove_consumers(self, user_emails):
        return self._remove_permissions(user_emails, pb.PermissionType.Consumer)

    def remove_sensitive_consumers(self, user_emails):
        """Remove sensitive consumer/consumers from project.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            Project: An existing project with the latest information.

        Typical example:
            project.remove_sensitive_consumers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#project-permission-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.SensitiveConsumer)

    def remove_viewers(self, user_emails):
        """Remove viewer/viewers from project.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            Project: An existing project with the latest information.

        Typical example:
            project.remove_viewers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#project-permission-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.Viewer)

    @property
    def owners(self):
        request = pb.GetOwnersRequest(resource_id=self._project.id)
        response = self._stub.GetProjectOwners(request)
        return [User(owner) for owner in response.owners]

    def _add_permissions(self, user_emails, permission):
        request = pb.ProjectPermissionRequest()
        request.project.CopyFrom(self._project)
        request.user_emails.extend(user_emails)
        request.permission = permission
        self._stub.AddProjectPermission(request)
        return self

    def _remove_permissions(self, user_emails, permission):
        request = pb.ProjectPermissionRequest()
        request.project.CopyFrom(self._project)
        request.user_emails.extend(user_emails)
        request.permission = permission
        self._stub.RemoveProjectPermission(request)
        return self

    def _refresh(self):
        request = pb.GetProjectRequest()
        request.project_name = self._project.name
        response = self._stub.GetProject(request)
        self._project = response.project

    def request_access(self, access_type, reason):
        request = pb.SubmitPendingPermissionRequest()
        request.resource_id = self._project.id
        request.permission = AccessType.to_proto_permission(access_type)
        request.reason = reason
        response = self._stub.SubmitPendingProjectPermission(request)
        return response.permission_id

    @property
    def current_permission(self):
        request = pb.GetActivePermissionRequest(resource_id=self._project.id)
        response = self._stub.GetActiveProjectPermission(request)
        return AccessType.from_proto_active_permission(response.permission)

    def __repr__(self):
        return Utils.pretty_print_proto(self._project)

    def __str__(self):
        return (
            f"Project name    : {self.name} \n"
            f"Description     : {self.description} \n"
            f"Secret          : {self.secret} \n"
            f"Locked          : {self.locked} \n"
            f"Author            \n{Utils.output_indent_spacing(str(self.author), '  ')}"
            f"Created         : {Utils.proto_to_dict(self._project).get('created_date_time')} \n"
            f"Last updated    : {Utils.proto_to_dict(self._project).get('last_update_date_time')} \n"
        )



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..utils import Utils


class PersonalAccessToken:
    def __init__(self, stub, pat):
        self._pat = pat
        self._stub = stub

    def revoke(self):
        request = pb.TokenRequest()
        request.token_id = self._pat.id
        self._stub.RevokeToken(request)

    @property
    def id(self):
        return self._pat.id

    @property
    def name(self):
        return self._pat.name

    @property
    def description(self):
        return self._pat.description

    @property
    def expiry_date(self):
        return Utils.timestamp_to_string(self._pat.expiry_date)

    @property
    def last_used(self):
        return Utils.timestamp_to_string(self._pat.last_used)

    @property
    def creation_time(self):
        return Utils.timestamp_to_string(self._pat.creation_time)

    def __repr__(self):
        return Utils.pretty_print_proto(self._pat)

    def __str__(self):
        return (
            f"Name          : {self.name} \n"
            f"Description   : {self.description} \n"
            f"Creation time : {self.creation_time} \n"
        )
